@@ -1,16 +1,19 @@
+import copy
 import logging
-import os
 
 import six
 
 from sevenbridges.errors import SbgError
-from sevenbridges.http.client import HttpClient
-from sevenbridges.meta.data import DataContainer
 from sevenbridges.meta.fields import Field
+from sevenbridges.meta.data import DataContainer
+from sevenbridges.meta.transformer import Transform
 
 logger = logging.getLogger(__name__)
 
 
+DEFAULT_LIMIT = 100
+
+
 # noinspection PyProtectedMember
 class ResourceMeta(type):
     """
@@ -23,7 +26,7 @@ class ResourceMeta(type):
     resource fields.
     """
 
-    def __new__(cls, name, bases, dct):
+    def __new__(mcs, name, bases, dct):
         # Attach fields object fo resource instance.
         fields = {}
         for k, v in dct.items():
@@ -39,23 +42,51 @@ class ResourceMeta(type):
         if '__init__' not in dct:
             def init(self, **kwargs):
                 self._api = kwargs.pop('api', None)
-                try:
-                    urls = self._URL
-                except AttributeError:
-                    urls = None
-                self._data = DataContainer(urls=urls, api=self._api)
+                urls = getattr(self, '_URL', None)
+                self._data = DataContainer(
+                    urls=urls, api=self._api, parent=self
+                )
                 self._dirty = {}
                 for k, v in kwargs.items():
                     if k in fields:
                         value = fields[k].validate(v)
                         self._data[k] = value
 
+                self._old = copy.deepcopy(self._data.data)
+
+            def _data_diff(d1, d2):
+                data = {}
+                new_keys = d2.keys() if isinstance(d2, dict) else []
+                old_keys = d1.keys() if isinstance(d1, dict) else []
+                for key in new_keys:
+                    if key not in old_keys:
+                        data[key] = d2[key]
+                    elif isinstance(d1[key], dict):
+                        inner_diff = _data_diff(d1[key], d2[key])
+                        if inner_diff:
+                            data[key] = inner_diff
+                    elif d1[key] != d2[key]:
+                        data[key] = d2[key]
+                return data
+
             # get modified data from the instance
             def modified_data(self):
+                metadata = copy.deepcopy(self._dirty.get('metadata'))
+                difference = _data_diff(self._old, self._data.data)
+                self._dirty.update(difference)
+                if metadata:
+                    # File metadata specific patch, otherwise the diff will
+                    # only return changed parameters even when metadata needs
+                    # to be replaced
+                    self._dirty['metadata'] = metadata
                 return self._dirty
 
             def equals(self, other):
-                return type(self) == type(other) and self._data == other._data
+                if not hasattr(other, '__class__'):
+                    return False
+                if not self.__class__ == other.__class__:
+                    return False
+                return self is other or self._data == other._data
 
             def deepcopy(self):
                 return self.__class__(api=self._api, **self._data.data)
@@ -73,12 +104,9 @@ class ResourceMeta(type):
             dct['deepcopy'] = deepcopy
             dct['_modified_data'] = modified_data
 
-        return type.__new__(cls, name, bases, dct)
+        return type.__new__(mcs, name, bases, dct)
 
     def __get__(cls, obj, objtype=None):
-        # SPHINX_DOC part is for generating documentation
-        if not isinstance(obj, HttpClient) and not os.environ['SPHINX_DOC']:
-            raise SbgError(message='Improperly configured client!')
         if obj is None:
             return cls
         cls._API = obj
@@ -106,19 +134,24 @@ class Resource(six.with_metaclass(ResourceMeta)):
         from sevenbridges.models.link import Link
         from sevenbridges.meta.collection import Collection
 
-        #: :type: _HttpClient
         api = kwargs.pop('api', cls._API)
         url = kwargs.pop('url')
+
+        # Check for valid limit value
+        if kwargs.get('limit') is not None and kwargs['limit'] <= 0:
+            kwargs['limit'] = DEFAULT_LIMIT
+
         extra = {'resource': cls.__name__, 'query': kwargs}
         logger.info('Querying {} resource'.format(cls), extra=extra)
         response = api.get(url=url, params=kwargs)
         data = response.json()
         total = response.headers['x-total-matching-query']
-        projects = [cls(api=api, **item) for item in data['items']]
+
+        items = [cls(api=api, **item) for item in data['items']]
         links = [Link(**link) for link in data['links']]
         href = data['href']
         return Collection(
-            resource=cls, href=href, total=total, items=projects,
+            resource=cls, href=href, total=total, items=items,
             links=links, api=api
         )
 
@@ -130,8 +163,7 @@ class Resource(six.with_metaclass(ResourceMeta)):
         :param api: sevenbridges Api instance.
         :return: Resource object.
         """
-        if not id:
-            raise SbgError('Invalid id value!')
+        id = Transform.to_resource(id)
         api = api if api else cls._API
         if 'get' in cls._URL:
             extra = {'resource': cls.__name__, 'query': {'id': id}}
@@ -139,7 +171,7 @@ class Resource(six.with_metaclass(ResourceMeta)):
             resource = api.get(url=cls._URL['get'].format(id=id)).json()
             return cls(api=api, **resource)
         else:
-            raise SbgError('Unable to fetch resource!')
+            raise SbgError('Unable to retrieve resource!')
 
     def delete(self):
         """
@@ -158,19 +190,48 @@ class Resource(six.with_metaclass(ResourceMeta)):
         Refreshes the resource with the data from the server.
         """
         try:
-            extra = {'resource': self.__class__.__name__, 'query': {
-                'id': self.id}}
-            logger.info('Reloading {} resource.'.format(self), extra=extra)
-            data = self._api.get(self.href, append_base=False).json()
-            resource = self.__class__(api=self._api, **data)
-        except Exception:
-            try:
+            if hasattr(self, 'href'):
+                data = self._api.get(self.href, append_base=False).json()
+                resource = self.__class__(api=self._api, **data)
+            elif hasattr(self, 'id') and hasattr(self, '_URL') and \
+                    'get' in self._URL:
                 data = self._api.get(
                     self._URL['get'].format(id=self.id)).json()
                 resource = self.__class__(api=self._api, **data)
-            except Exception:
-                raise SbgError('Resource can not be refreshed!')
+            else:
+                raise SbgError(
+                    'Resource can not be refreshed, "id" property not set or '
+                    'retrieval for this resource is not available.'
+                )
+            query = {'id': self.id} if hasattr(self, 'id') else {}
+            extra = {'resource': self.__class__.__name__, 'query': query}
+            logger.info('Reloading {} resource.'.format(self), extra=extra)
+        except Exception as e:
+            raise SbgError(
+                'Resource can not be refreshed due to an error: {}'
+                .format(six.text_type(e))
+            )
 
         self._data = resource._data
         self._dirty = resource._dirty
+        self.update_old()
         return self
+
+    def update_old(self):
+        self._old = copy.deepcopy(self._data.data)
+
+    def field(self, name):
+        """
+        Return field value if it's set
+        :param name: Field name
+        :return: Field value or None
+        """
+        return self._data.data.get(name, None)
+
+    def _set(self, key, value):
+        """
+        Set property value in internal storage
+        :param key: Property name
+        :param value: Property value
+        """
+        self._data.data[key] = value
