@@ -1,8 +1,8 @@
 import io
-import math
 import os
-import threading
 import time
+import logging
+import threading
 
 import six
 
@@ -12,6 +12,8 @@ from sevenbridges.http.client import generate_session
 from sevenbridges.models.enums import PartSize, TransferState
 from sevenbridges.transfer.utils import Progress, total_parts
 
+logger = logging.getLogger(__name__)
+
 
 def _get_part_url(api, url, upload, part):
     """
@@ -73,10 +75,11 @@ def _submit_part(session, url, part, timeout):
         response = session.put(url, data=part, timeout=timeout)
         return response.headers.get('etag').strip('"')
     except Exception as e:
+        logger.debug(
+            'Failed to submit part due to an error: %s', six.text_type(e)
+        )
         raise SbgError(
-            'Failed to submit the part. Reason: {}'.format(
-                six.text_type(e)
-            )
+            'Failed to submit the part. Reason: {}'.format(six.text_type(e))
         )
 
 
@@ -209,10 +212,9 @@ class Upload(threading.Thread):
         'upload_complete': '/upload/multipart/{upload_id}/complete'
     }
 
-    def __init__(self, file_path, project, file_name=None, overwrite=False,
-                 part_size=PartSize.UPLOAD_MINIMUM_PART_SIZE, retry_count=5,
-                 timeout=60, api=None):
-
+    def __init__(self, file_path, project=None, parent=None, file_name=None,
+                 overwrite=False, part_size=None, retry_count=5, timeout=60,
+                 api=None):
         """
         Multipart File uploader.
 
@@ -228,8 +230,13 @@ class Upload(threading.Thread):
         threading.Thread.__init__(self)
         self.daemon = True
 
-        if not project:
-            raise SbgError('Project identifier is not valid.')
+        if not project and not parent:
+            raise SbgError('Project or parent identifier is required.')
+
+        if project and parent:
+            raise SbgError(
+                'Project and parent identifiers are mutually exclusive.'
+            )
 
         if not file_path:
             raise SbgError('File path is not valid.')
@@ -246,21 +253,15 @@ class Upload(threading.Thread):
         else:
             self._file_name = file_name
 
-        if part_size and part_size < PartSize.UPLOAD_MINIMUM_PART_SIZE:
-            raise SbgError(
-                'Chunk size is too small! Minimum get_parts size is {}'.format(
-                    PartSize.UPLOAD_MINIMUM_PART_SIZE)
-            )
-        self._part_size = part_size
+        self._part_size = part_size  # or PartSize.UPLOAD_MINIMUM_PART_SIZE
         self._project = project
+        self._parent = parent
         self._file_path = file_path
         self._file_size = os.path.getsize(self._file_path)
         if self._file_size == 0:
             raise SbgError('File size must not be 0.')
 
         self._verify_file_size()
-        self._verify_part_size()
-        self._verify_part_number()
 
         self._overwrite = overwrite
         self._retry = retry_count
@@ -276,7 +277,12 @@ class Upload(threading.Thread):
         self._stop_signal = False
         self._result = None
 
-        self.session = generate_session(self._api.session.proxies)
+        self.session = generate_session(
+            self._api.pool_connections,
+            self._api.pool_maxsize,
+            self._api.pool_block,
+            self._api.session.proxies
+        )
 
     def __repr__(self):
         return six.text_type(
@@ -286,30 +292,6 @@ class Upload(threading.Thread):
     def result(self):
         return self._result
 
-    def _verify_part_number(self):
-        """
-        Verifies that the total number of parts is smaller then 10^5 which
-        is the maximum number of parts.
-        """
-        total = int(math.ceil(self._file_size / self._part_size))
-        if total > PartSize.MAXIMUM_TOTAL_PARTS:
-            self._status = TransferState.FAILED
-            raise SbgError(
-                'Total parts = {}. Maximum number of parts is {}'.format(
-                    total, PartSize.MAXIMUM_TOTAL_PARTS)
-            )
-
-    def _verify_part_size(self):
-        """
-        Verifies that the part size is smaller then the maximum part size
-        which is 5GB.
-        """
-        if self._part_size > PartSize.MAXIMUM_UPLOAD_SIZE:
-            self._status = TransferState.FAILED
-            raise SbgError('Part size = {}b. Maximum part size is {}b'.format(
-                self._part_size, PartSize.MAXIMUM_UPLOAD_SIZE)
-            )
-
     def _verify_file_size(self):
         """
         Verifies that the file is smaller then 5TB which is the maximum
@@ -322,18 +304,21 @@ class Upload(threading.Thread):
             )
 
     def _initialize_upload(self):
-
         """
         Initialized the upload on the API server by submitting the information
         about the project, the file name, file size and the part size that is
         going to be used during multipart upload.
         """
         init_data = {
-            'project': self._project,
             'name': self._file_name,
-            'part_size': self._part_size,
-            'size': self._file_size
+            'size': self._file_size,
         }
+        if self._part_size:
+            init_data['part_size'] = self._part_size
+        if self._project:
+            init_data['project'] = self._project
+        elif self._parent:
+            init_data['parent'] = self._parent
 
         init_params = {}
         if self._overwrite:
@@ -343,7 +328,12 @@ class Upload(threading.Thread):
             response = self._api.post(
                 self._URL['upload_init'], data=init_data, params=init_params
             )
-            self._upload_id = response.json()['upload_id']
+            data = response.json()
+            self._upload_id = data['upload_id']
+            part_size = data.get('part_size')
+            if part_size and part_size != self._part_size:
+                logger.debug('Part size optimized by server to %s', part_size)
+                self._part_size = part_size
         except SbgError as e:
             self._status = TransferState.FAILED
             raise SbgError(
@@ -493,31 +483,29 @@ class Upload(threading.Thread):
 
         # Opens the file for reading in binary mode.
         try:
-            fp = io.open(self._file_path, mode='rb')
+            with io.open(self._file_path, mode='rb') as fp:
+                # Creates a partitioned file
+                parted_file = UPartedFile(
+                    fp, self._file_size, self._part_size, self._upload_id,
+                    self._retry, self._timeout, self.session, self._api
+                )
+
+                # Iterates over parts and submits them for upload.
+                for _ in parted_file:
+                    if self._stop_signal:
+                        return
+                    self._running.wait()
+                    self._bytes_done += self._part_size
+                    # If the progress callback is set we need to provide a
+                    # progress object for it.
+                    if self._progress_callback:
+                        progress = Progress(
+                            parted_file.total, parted_file.total_submitted,
+                            self._bytes_done, self._file_size, self.duration
+                        )
+                        self._progress_callback(progress)
         except IOError:
             raise SbgError('Unable to open file {}'.format(self._file_path))
-
-        # Creates a partitioned file
-        parted_file = UPartedFile(
-            fp, self._file_size, self._part_size, self._upload_id,
-            self._retry, self._timeout, self.session, self._api
-        )
-
-        # Iterates over parts and submits them for upload.
-        try:
-            for _ in parted_file:
-                if self._stop_signal:
-                    return
-                self._running.wait()
-                self._bytes_done += self._part_size
-                # If the progress callback is set we need to provide a
-                # progress object for it.
-                if self._progress_callback:
-                    progress = Progress(
-                        parted_file.total, parted_file.total_submitted,
-                        self._bytes_done, self._file_size, self.duration
-                    )
-                    self._progress_callback(progress)
         except Exception as e:
             # If the errorback callback is set call it with status
             self._status = TransferState.FAILED
